#pragma once

#include <list>
#include <memory>

#include <amdis/Boundary.hpp>
#include <amdis/LocalAssemblerBase.hpp>
#include <amdis/utility/TreeData.hpp>

namespace AMDiS
{
  namespace tag
  {
    template <class E> struct element_operator { using type = E; };
    template <class I> struct intersection_operator { using type = I; };
    template <class I> struct boundary_operator { using type = I; BoundaryType id; };
  }

  template <class GridView>
  class OperatorLists
  {
    using Element = typename GridView::template Codim<0>::Entity;
    using Intersection = typename GridView::Intersection;

    template <class OperatorType>
    struct DataElement
    {
      std::shared_ptr<OperatorType> op;
      BoundaryType b = {0};
    };

    /// Lists of \ref DataElement on the Element, BoundaryIntersction, and InteriorIntersections
    template <class... Nodes>
    struct Data
    {
      /// Return whether any operators are stored on the node
      bool empty() const
      {
        return element_.empty() && boundary_.empty() && intersection_.empty();
      }

      /// Test whether to assemble on the node
      bool doAssemble() const
      {
        return !assembled_ || changing_;
      }

      operator bool() const
      {
        return doAssemble() && !empty();
      }

      /// Bind all operators to the grid element and geometry
      template <class Geo>
      void bind(Element const& elem, Geo const& geo)
      {
        for (auto& scaled : element_) scaled.op->bind(elem,geo);
        for (auto& scaled : boundary_) scaled.op->bind(elem,geo);
        for (auto& scaled : intersection_) scaled.op->bind(elem,geo);
      }

      /// Unbind all operators from the element
      void unbind()
      {
        for (auto& scaled : element_) scaled.op->unbind();
        for (auto& scaled : boundary_) scaled.op->unbind();
        for (auto& scaled : intersection_) scaled.op->unbind();
        assembled_ = true;
      }


      template <class... Args>
      void push(tag::element_operator<Element>, Args&&... args)
      {
        element_.push_back({std::forward<Args>(args)...});
      }

      template <class... Args>
      void push(tag::boundary_operator<Intersection> b, Args&&... args)
      {
        boundary_.push_back({std::forward<Args>(args)..., b.id});
      }

      template <class... Args>
      void push(tag::intersection_operator<Intersection>, Args&&... args)
      {
        intersection_.push_back({std::forward<Args>(args)...});
      }


      auto&       onElement()       { return element_; }
      auto const& onElement() const { return element_; }

      auto&       onBoundary()       { return boundary_; }
      auto const& onBoundary() const { return boundary_; }

      auto&       onIntersection()       { return intersection_; }
      auto const& onIntersection() const { return intersection_; }

    private:
      /// The type of local operators associated with grid elements
      using ElementOperator = LocalAssemblerBase<Element, Nodes...>;

      /// The type of local operators associated with grid intersections
      using IntersectionOperator = LocalAssemblerBase<Intersection, Nodes...>;

      /// List of operators to be assembled on grid elements
      std::vector<DataElement<ElementOperator>> element_;

      /// List of operators to be assembled on boundary intersections
      std::vector<DataElement<IntersectionOperator>> boundary_;

      /// List of operators to be assembled on interior intersections
      std::vector<DataElement<IntersectionOperator>> intersection_;

      /// if false, do reassemble of all operators
      bool assembled_ = false;

      /// if true, or \ref assembled false, do reassemble of all operators
      bool changing_ = true;
    };

  public:

    /// List of operators associated with matrix blocks (RowNode, ColNode)
    template <class RowNode, class ColNode>
    using MatData = Data<RowNode, ColNode>;

    /// List of operators associated with vector blocks [Node]
    template <class Node>
    using VecData = Data<Node>;
  };


  template <class RowBasis, class ColBasis>
  using MatrixOperators = MatrixData<RowBasis, ColBasis, OperatorLists<typename RowBasis::GridView>::template MatData>;

  template <class GlobalBasis>
  using VectorOperators = VectorData<GlobalBasis, OperatorLists<typename GlobalBasis::GridView>::template VecData>;

} // end namespace AMDiS
