#pragma once

#include <list>
#include <memory>
#include <string>
#include <vector>

#include <Eigen/SparseCore>

#include <dune/common/timer.hh>

#include <amdis/Output.hpp>
#include <amdis/linear_algebra/Common.hpp>
#include <amdis/linear_algebra/DOFMatrixBase.hpp>

namespace AMDiS
{
  /// \brief The basic container that stores a base matrix and a corresponding
  /// row/column feSpace.
  template <class ValueType, int Orientation = Eigen::RowMajor>
  class EigenMatrix
  {
  public:
    /// The matrix type of the underlying base matrix
    using BaseMatrix = Eigen::SparseMatrix<ValueType, Orientation>;

    /// The type of the elements of the DOFMatrix
    using value_type = ValueType;

    /// The index/size - type
    using size_type = typename BaseMatrix::Index;

  public:
    /// Constructor. Constructs new BaseMatrix.
    EigenMatrix() = default;

    /// Return a reference to the data-matrix \ref matrix
    BaseMatrix& matrix()
    {
      return matrix_;
    }

    /// Return a reference to the data-matrix \ref matrix
    BaseMatrix const& matrix() const
    {
      return matrix_;
    }


    /// \brief Returns an update-proxy of the inserter, to inserte/update a value at
    /// position (\p r, \p c) in the matrix. Need an insertionMode inserter, that can
    /// be created by \ref init and must be closed by \ref finish after insertion.
    void insert(size_type r, size_type c, value_type const& value)
    {
      test_exit_dbg(r < matrix_.rows() && c < matrix_.cols(),
          "Indices out of range [0,{})x[0,{})", matrix_.rows(), matrix_.cols());
      tripletList_.emplace_back(r, c, value);
    }


    /// Create inserter. Assumes that no inserter is currently active on this matrix.
    template <class RowBasis, class ColBasis>
    void init(RowBasis const& rowBasis, ColBasis const& colBasis,
              bool prepareForInsertion)
    {
      matrix_.resize(rowBasis.dimension(), colBasis.dimension());
      if (prepareForInsertion) {
        matrix_.setZero(); // maybe not necessary
      }
    }

    /// Delete inserter -> finish insertion. Must be called in order to fill the
    /// final construction of the matrix.
    void finish()
    {
      matrix_.setFromTriplets(tripletList_.begin(), tripletList_.end());
      matrix_.makeCompressed();

      tripletList_.clear(); // NOTE: this does not free the memory
    }

    /// Return the number of nonzeros in the matrix
    std::size_t nnz() const
    {
      return matrix_.nonZeros();
    }

    /// \brief Deletes all rows with \p dirichletNodes != 0 and if \p apply is true, adds
    /// a one on the diagonal of the matrix.
    auto applyDirichletBC(std::vector<char> const& dirichletNodes)
    {
      for (std::size_t i = 0; i < dirichletNodes.size(); ++i) {
        if (dirichletNodes[i] == 1)
          clearDirichletRow(i, int_<Orientation>);
      }
      return std::list<Triplet<value_type>>{};
    }


  protected:
    void clearDirichletRow(size_type idx, int_t<Eigen::ColMajor>)
    {
      for (size_type k = 0; k < matrix_.outerSize(); ++k) {
        for (typename BaseMatrix::InnerIterator it(matrix_, k); it; ++it) {
          if (it.row() == idx) {
            it.valueRef() = (it.row() == it.col() ? value_type(1) : value_type(0));
            break;
          }
        }
      }
    }

    void clearDirichletRow(size_type idx, int_t<Eigen::RowMajor>)
    {
      Eigen::SparseVector<ValueType> unit_vector(matrix_.cols());
      unit_vector.coeffRef(idx) = value_type(1);
      matrix_.row(idx) = unit_vector;
    }

  private:
    /// The data-matrix
    BaseMatrix matrix_;

    /// A list of row,col indices and values
    std::vector<Eigen::Triplet<ValueType, Eigen::Index>> tripletList_;
  };

  template <class RowBasisType, class ColBasisType,
            class T = double, int O = Eigen::RowMajor>
  using DOFMatrix = DOFMatrixBase<RowBasisType, ColBasisType, EigenMatrix<T, O>>;

} // end namespace AMDiS
